/*****************************************************************************
 * Copyright (C) 2006 Pter Ekler, Zsolt Pszota
 * ---------------------------------------------------------------------------
 * This file is part of SymTracker
 *
 * SymTracker is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SymTracker is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SymTracker; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

// INCLUDE FILES
#include  "STTTransferSocketInitializator.h"

// Constructors and destructor ------------------------------------------------
CSTTTransferSocketInitializator::CSTTTransferSocketInitializator(MSTTSocketCollector* aSocketCollector)
:CActive(EPriorityNormal),iSocketCollector(aSocketCollector)
{
}

void CSTTTransferSocketInitializator::ConstructL(TPtrC aServerName, TInt aServicePort, TPtrC8 aConnectionId)
{
	iState = ETSInitializing;
	
	iSocket = new (ELeave) RSocket;

	iServerName			= HBufC::NewL(160);
	(iServerName->Des()).Copy(aServerName);
	iConnectionId		= HBufC8::NewL(160);
	(iConnectionId->Des()).Copy(aConnectionId);
	iServicePort	= aServicePort;
	iIpSet			= EFalse;
	iPortSet		= EFalse;
	User::LeaveIfError(iSockServ.Connect());

	iNextSibling	= NULL;

	CActiveScheduler::Add(this);

	TransferConnectL();
}

CSTTTransferSocketInitializator::~CSTTTransferSocketInitializator()
{
	Cancel();

	TransferDisconnect();

	delete iServerName;
	delete iConnectionId;
	
	if (iSocketPassed)
		delete iSocket;
	
	iSockServ.Close();
}

CSTTTransferSocketInitializator* CSTTTransferSocketInitializator::NewL(
		TPtrC aServerName, TInt aServicePort, TPtrC8 aConnectionId, MSTTSocketCollector* aSocketCollector)
{
	CSTTTransferSocketInitializator * self=new(ELeave) CSTTTransferSocketInitializator(aSocketCollector);
	CleanupStack::PushL(self);
	self->ConstructL(aServerName, aServicePort, aConnectionId);
	CleanupStack::Pop();
	return self;
}

// Getter and setter methods --------------------------------------------------

TTransferSocketStates CSTTTransferSocketInitializator::GetState()
{
	return iState;
}

CSTTTransferSocketInitializator* CSTTTransferSocketInitializator::GetNextSibling()
{
	return iNextSibling;
}

void CSTTTransferSocketInitializator::SetNextSibling(CSTTTransferSocketInitializator* aNewNextSibling)
{
	iNextSibling = aNewNextSibling;
}

// Communication methods ------------------------------------------------------

void CSTTTransferSocketInitializator::TransferConnectL()
{
	
	if (iState != ETSInitializing) return;

	TInetAddr addr;
	addr.Input((iServerName->Des()));
	addr.SetPort(iServicePort);

	User::LeaveIfError(iSocket->Open(iSockServ, KAfInet, KSockStream, KProtocolInetTcp));

	iSocket->Connect(addr,iStatus);	// opening socket
	SetActive();
	iState = ETSConnecting;
	
}

void CSTTTransferSocketInitializator::TransferDisconnect()
{
	if (iState==ETSInitializing) return;

	Cancel();
	if (iState!=ETSFinished)
		iSocket->Close();				
    iState = ETSInitializing;
}


void CSTTTransferSocketInitializator::SendInitialization()
{
	if (iState!=ETSConnecting) return;

	iState = ETSSendingInitialization;

	iSendingBuffer.Copy(_L8("#paid#"));
	iSendingBuffer.Append(iConnectionId->Des());
	iSendingBuffer.Append(_L8("\n"));
	iSocket->Write(iSendingBuffer,iStatus);
	SetActive();
}

void CSTTTransferSocketInitializator::StartWaitingForData()
{
	iTemporaryBuffer.Copy(_L8(""));
	WaitForData();
}

void CSTTTransferSocketInitializator::WaitForData()
{
	iState = ETSWaitingForData;
	iSocket->RecvOneOrMore(iReceivingBuffer, 0, iStatus, iRecievedLength);
	SetActive();
}

void CSTTTransferSocketInitializator::ProcessData()
{
	if ((iTemporaryBuffer.Length()==iTemporaryBuffer.MaxLength()) || // buffer exceed
		(iReceivingBuffer.Length()==0)) // nothing has arrived, something is wrong
		{ Error(); return; }

	if (iReceivingBuffer[0]!='\n') // not a newline
		iTemporaryBuffer.Append(iReceivingBuffer); // append
	else
	{
		if (iTemporaryBuffer.Left(6).Compare(_L8("#paip#"))==0)
		{
			// partner ip address
			TBuf<100> buf; // 8->16 conversion
			buf.Copy(iTemporaryBuffer.Right(iTemporaryBuffer.Length()-6));

			if (iPartnerAddress.Input(buf)!=KErrNone)
				{ Error(); return; }
			iIpSet = ETrue;
		}
		else if (iTemporaryBuffer.Left(6).Compare(_L8("#papo#"))==0)
		{
			// partner port
			TInt tempport;
			if (TLex8(iTemporaryBuffer.Right(iTemporaryBuffer.Length()-6)).Val(tempport)!=KErrNone)
				{ Error(); return; }

			iPartnerAddress.SetPort(tempport);
			iPortSet = ETrue;
		}

		if (iIpSet && iPortSet)
		{
			FinishInitialization();
			return;
		}

		iTemporaryBuffer.SetLength(0);
	}

	WaitForData();
}

void CSTTTransferSocketInitializator::FinishInitialization()
{
	iSocketCollector->Accept(iSocket ,iPartnerAddress);
	iSocketPassed = ETrue;
	iState = ETSFinished;	
}

// Active object methods ------------------------------------------------------

void CSTTTransferSocketInitializator::DoCancel()
{
	iSocket->CancelAll();
	// cancel all asychronous stuff
}

void CSTTTransferSocketInitializator::RunL()
{
    switch (iStatus.Int())
    {
        case KErrNone:
			// TODO: Handle succeed events 
			switch (iState)
			{
				case ETSConnecting: // connection succeed
					SendInitialization();
					break;
				case ETSSendingInitialization: // initialization sent
					StartWaitingForData();
					break;
				case ETSWaitingForData:
					ProcessData();
					break;
				default:
					break;
			}
			break;
		default:
			// TODO: Handle errors 
			switch (iState)
			{
				default:
					Error();
					break;
			}
			break;
	}

}

void CSTTTransferSocketInitializator::Error()
{
	TransferDisconnect();
	iState = ETSFinished;
	return;
}
