/*****************************************************************************
 * Copyright (C) 2006-2007 Imre Kelnyi
 *-------------------------------------------------------------------
 * This file is part of SymTorrent
 *
 * SymTorrent is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SymTorrent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Symella; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* ============================================================================
 *  Name     : CSTPiece from STPiece.cpp
 *  Part of  : SymTorrent
 *  Created  : 10.02.2006 by Imre Kelnyi
 * ============================================================================
 */

// INCLUDE FILES
#include "STPiece.h"
#include "STGlobal.h"
#include "SymTorrentEngineLog.h"
#include "STFile.h"
#include "STFileManager.h"
#include "hash.h"

#define LOG iTorrent.Log()

// ================= MEMBER FUNCTIONS =======================

CSTPiece::CSTPiece(CSTTorrent& aTorrent, TBuf8<20>& aHash, TInt aTotalSize)
 : iTorrent(aTorrent), iHash(aHash), iTotalSize(aTotalSize)
{
}

CSTPiece::~CSTPiece()
{
	iFiles.Reset();
}

void CSTPiece::IncreaseDownloadedSizeL(TInt aSize, CSTPeer* aPeer)
{
	iDownloadedSize += aSize;
	iTorrent.UpdateBytesDownloadedL(aSize);
	
	if (Remaining() == 0)
	{
		LogPieceIndexL();
		LWRITELN(LOG, _L("Downloading piece completed!"));
		if (CheckHashL())
		{
			LogPieceIndexL();
			LWRITELN(LOG, _L("Hash OK!"));
			SetFilesDownloaded();
			iTorrent.PieceDownloadedL(this, aPeer);
		}
		else
		{
			LogPieceIndexL();
			LWRITELN(LOG, _L("Hash FAILED!"));
			iDownloadedSize = 0;
			iTorrent.PieceHashFailedL(this);
		}		
	}	
}

TInt CSTPiece::AppendBlockL(const TDesC8& aBlock, CSTPeer* aPeer)
{
	TInt blockPosition = 0;

	while (blockPosition < aBlock.Length())
	{
		CSTFile* file = NULL;
		TInt filePosition = 0;

		GetFilePosition(iDownloadedSize, &file, filePosition);

		if (file)
		{
			TInt res = KErrNone;
			TInt fileBlockLength = file->Size() - filePosition;

			LogPieceIndexL();
			LWRITE(LOG, _L("Appending block to "));
			LWRITELN(LOG, file->Path());

			if (fileBlockLength <= (aBlock.Length() - blockPosition))	// reached the end of a file
			{
				LogPieceIndexL();
				LWRITE(LOG, _L("File complete: "));
				LWRITELN(LOG, file->Path());

				res = FileManager().WriteL(file, 
					filePosition, aBlock.Mid(blockPosition, fileBlockLength));
				if (res != KErrNone)
					return res;

				FileManager().Close(file);

				blockPosition += fileBlockLength;
				iDownloadedSize += fileBlockLength;				
			}
			else // the file isn't finished yet
			{
				res = FileManager().WriteL(file, 
					filePosition, aBlock.Right(aBlock.Length() - blockPosition));											

				if (res != KErrNone)
					return res;

				iDownloadedSize += (aBlock.Length() - blockPosition);
				//blockPosition = aBlock.Length();
				break;
			}
		}
		else
			//User::Panic(KLitSymTorrent, ESTPanGetFilePositionFailed);
			return KErrGeneral;
	}

	iTorrent.UpdateBytesDownloadedL(aBlock.Length());

	//LOG->WriteL(_L("[Piece] Append succeeded, remaining bytes: "));
	//LOG->WriteLineL(Remaining());

	if (Remaining() == 0)
	{
		LogPieceIndexL();
		LWRITELN(LOG, _L("Downloading piece completed!"));
		if (CheckHashL())
		{
			LogPieceIndexL();
			LWRITELN(LOG, _L("Hash OK!"));
			SetFilesDownloaded();
			iTorrent.PieceDownloadedL(this, aPeer);
		}
		else
		{
			LogPieceIndexL();
			LWRITELN(LOG, _L("Hash FAILED!"));
			iDownloadedSize = 0;
			iTorrent.PieceHashFailedL(this);
		}		
	}
	else
		if (Remaining() < 0)
			return KErrGeneral;

	return KErrNone;
}

void CSTPiece::SetDownloadedSize(TInt aDownloadedSize)
{
	iDownloadedSize = aDownloadedSize;
	
	// TODO replace following code with assertation and panic
	if (iDownloadedSize >= iTotalSize)
		iDownloadedSize = 0;
}

void CSTPiece::SetDownloaded(TBool aDownloaded)
{
	if (aDownloaded)
	{
		iDownloadedSize = iTotalSize;
		SetFilesDownloaded();
	}
	else
	{
		iDownloadedSize = 0;
	
		for (TInt i=0; i<iFiles.Count(); i++)
			iFiles[i].iFile.SetDownloaded(EFalse);
	}
}

void CSTPiece::SetFilesDownloaded() // !!! SHOULD BE REVISED !!!
{
	for (TInt i=0; i<iFiles.Count(); i++)
	{
		TBool downloaded = ETrue;
		for (TInt j=0; j<iTorrent.PieceCount(); j++)
		{
			if ((iTorrent.Piece(j)->HasFile(&iFiles[i].iFile)) &&
				!(iTorrent.Piece(j)->IsDownloaded()))
			{
				downloaded = EFalse;
				break;
			}						
		}
		
		if (downloaded)
		{	
			LWRITE(LOG, _L("[File] "));
			LWRITE(LOG, iFiles[i].iFile.Path());
			LWRITELN(LOG, _L(" downloaded"));
			iFiles[i].iFile.SetDownloaded();
			FileManager().Close(&(iFiles[i].iFile));
		}			
	}
}


void CSTPiece::GetFilePosition(TInt aPositionInPiece, CSTFile** aFile, TInt& aPositionInFile)
{
	TInt pos = 0;

	for (TInt i=0; i<iFiles.Count(); i++)
	{
		if (aPositionInPiece < (pos + iFiles[i].iLength))
		{
			*aFile = &(iFiles[i].iFile);
			aPositionInFile = iFiles[i].iOffset + (aPositionInPiece - pos);
			return;
		}
		else
			pos += iFiles[i].iLength;
	}

	*aFile = NULL;
	return;
}


HBufC8* CSTPiece::GetBlockL(TInt aPosition, TInt aLength)
{	
	TInt filePos = 0;
	
	HBufC8* block = HBufC8::NewLC(aLength);
	TPtr8 blockPtr(block->Des());
	
	CSTFile* file = NULL;

	TInt i=0;
	TInt pos = 0;
	for (; i<iFiles.Count(); i++)
	{
		if (aPosition < (pos + iFiles[i].iLength))
		{
			file = &(iFiles[i].iFile);
			filePos = iFiles[i].iOffset + (aPosition - pos);
			break;
		}
		else
			pos += iFiles[i].iLength;
	}
	
	if (file == NULL)
	{
		CleanupStack::PopAndDestroy(); // block
		return NULL;		
	}
	
	while (aLength > 0)
	{
		if (TUint(aLength) >= (file->Size() - filePos))
		{
			HBufC8* buf = FileManager().ReadL(file, filePos, file->Size() - filePos);
			if (buf)
			{
				blockPtr.Append(*buf);
				delete buf;
				aLength -= (file->Size() - filePos);
			}
			else
				break;
		}
		else
		{
			HBufC8* buf = FileManager().ReadL(file, filePos, aLength);
			if (buf)
			{
				blockPtr.Append(*buf);
				aLength = 0;
				delete buf;				
			}
			
			break;		
		}
		
		if (i >= iFiles.Count()-1)
			break;
		
		file = &iFiles[++i].iFile;
		filePos = 0;
	}
	
	if (aLength != 0)
	{
		CleanupStack::PopAndDestroy(); // block
		return NULL;		
	}
		
	CleanupStack::Pop(); // block
	return block;	
}


void CSTPiece::AddFileFragmentL(CSTFile& aFile, TUint aOffset, TUint aLength)
{
	//LOG->WriteL(_L("[Piece "));
	//LOG->WriteL(Index());
	//LOG->WriteL(_L("] Adding file fragment of "));
	//LOG->WriteL(aFile.Path());
	//LOG->WriteL(_L(" Offset: "));
	//LOG->WriteL(aOffset);
	//LOG->WriteL(_L(" Length: "));
	//LOG->WriteLineL(aLength);

	User::LeaveIfError(iFiles.Append(TSTFileFragment(aFile, aOffset, aLength)));
}


TBool CSTPiece::CheckHashL()
{
//	if (Remaining() != 0)
//		return EFalse;
	
	TBool result = EFalse;
	
	HBufC8* piece = GetBlockL(0, TotalSize());
	if (piece)
	{
		CleanupStack::PushL(piece);
		CSHA1* sha1 = CSHA1::NewL();
		CleanupStack::PushL(sha1);
		if (sha1->Hash(*piece) == iHash)
			result = ETrue;
		
	//	LOG->WriteL(_L("[Piece] Hash from .torrent file   : "));
	//	LOG->WriteLineL(iHash);
		
	//	LOG->WriteL(_L("[Piece] Hash of the received piece: "));
	//	LOG->WriteLineL(sha1->Hash(*piece));

		CleanupStack::PopAndDestroy(); // sha1
		CleanupStack::PopAndDestroy(); // piece
	}
	
	return result;
}

TBool CSTPiece::HasFile(CSTFile* aFile)
{
	for (TInt i=0; i<iFiles.Count(); i++)
		if (&iFiles[i].iFile == aFile)
			return ETrue;
		
	return EFalse;
}

void CSTPiece::LogPieceIndexL()
{
	#ifdef LOG_TO_FILE
		TBuf<64> buf;
		buf.Format(_L("[Piece %d] "), Index());
		LOG->WriteL(buf);
	#endif
}

TBool CSTPiece::IsAllFilesSkipped() const 
{
	for (TInt i=0; i<iFiles.Count(); i++)
	{
		if (!iFiles[i].iFile.IsSkipped())
			return EFalse;
	}
	
	return ETrue;
}


