/*****************************************************************************
 * Copyright (C) 2006 Pter Ekler, Zsolt Pszota
 * ---------------------------------------------------------------------------
 * This file is part of SymTracker
 *
 * SymTracker is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SymTracker is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SymTracker; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

// INCLUDE FILES
#include <f32file.h>
#include "STTServiceSocket.h"
#include "STTTorrentPeerItem.h"
#include "STTTorrentListItem.h"
#include "SymTorrentTrackerManager.h"
#include "KiLogManager.h"
#include "STDefs.h"

// Constructors and destructor ------------------------------------------------

CSTTServiceSocket::CSTTServiceSocket(RSocket& aSocket, TInetAddr &aPartnerAddress, CSymTorrentTrackerManager* aTrackerManager):
	CActive(EPriorityNormal),iSocket(aSocket),iTrackerManager(aTrackerManager), iPartnerAddress(aPartnerAddress)
{
	iNextSibling	= NULL;
	iState			= ESSInitializing;

	iCurrentRequest.finished = EFalse;
	iCurrentRequest.requestedUrl = _L8("").AllocL();
}

void CSTTServiceSocket::ConstructL()
{
	iRecievedBuf		= HBufC8::NewL(16384);
	
	iSocket.RemoteName(iRemoteAddressFromSocket);

	CActiveScheduler::Add(this);
	ReceiveMessage();
	TBuf<100> buf;
	iPartnerAddress.Output(buf);
	buf.Append(_L(":"));
	buf.AppendNum(iPartnerAddress.Port());
	
	iLog = STLOG;
	
	iLog->WriteLineL(_L("[TRACKER] Partner address: "));
	iLog->WriteL(buf);
}

CSTTServiceSocket::~CSTTServiceSocket()
{
	TransferDisconnect();

	delete iCurrentRequest.requestedUrl;
	if (iCurrentRequest.infoHash!=NULL) delete iCurrentRequest.infoHash;
	if (iCurrentRequest.peerId!=NULL) delete iCurrentRequest.peerId;
	if (iCurrentRequest.compact!=NULL) delete iCurrentRequest.compact;
	if (iCurrentRequest.ip!=NULL) delete iCurrentRequest.ip;
	if (iCurrentRequest.key!=NULL) delete iCurrentRequest.key;
	if (iCurrentRequest.trackerId!=NULL) delete iCurrentRequest.trackerId;
	delete iRecievedBuf;
}

CSTTServiceSocket* CSTTServiceSocket::NewL(RSocket& aSocket, TInetAddr &aPartnerAddress, CSymTorrentTrackerManager* aTrackerManager)
{
	CSTTServiceSocket * self=new(ELeave) CSTTServiceSocket(aSocket, aPartnerAddress, aTrackerManager);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

// Getter and setter methods --------------------------------------------------

TBool CSTTServiceSocket::IsFinished()
{
	if (iState==ESSFinished)
		return ETrue;
	else
		return EFalse;
}

CSTTServiceSocket* CSTTServiceSocket::GetNextSibling()
{
	return iNextSibling;
}

void CSTTServiceSocket::SetNextSibling(CSTTServiceSocket* aNewNextSibling)
{
	iNextSibling = aNewNextSibling;
}

/* TORRENT TRACKER PROTOCOL ************************************************ */

// Communication methods ------------------------------------------------------
void CSTTServiceSocket::TransferDisconnect()
{
	if (iState!=ESSFinished)
	{
		Cancel();
		iSocket.Close();		
		iState = ESSFinished;
	}
}

void CSTTServiceSocket::ReceiveMessage()
{
	(iRecievedBuf->Des()).FillZ();

	ContinueReading();
}

void CSTTServiceSocket::ContinueReading()
{
	iState = ESSWaitingForIncoming;
	iTemporaryBuffer.FillZ();
	iSocket.RecvOneOrMore(iTemporaryBuffer, 0, iStatus, iRecievedLength);
	SetActive();

}

void CSTTServiceSocket::ReadIncoming()
{
	iState = ESSProcessIncoming;

	if (!ProcessIncomingL())
		ContinueReading();
	else
		SendRequestedInformationL();
}

TBool CSTTServiceSocket::ProcessIncomingL()
{
	TPtr8 pRecBuf(iRecievedBuf->Des());

	pRecBuf.Append(iTemporaryBuffer);
	iTemporaryBuffer.Copy(_L8(""));

	TBuf8<500> buf;
	TBuf8<500> tempbuf;
	TInt position;
	while ((position=pRecBuf.Locate('\n'))!=KErrNotFound)
	{
		buf = pRecBuf.Left(position);

		pRecBuf = pRecBuf.Right(pRecBuf.Length()-(position+1));

		// Format: GET /location HTTP/1.0 or GET /location HTTP/1.1
		if (buf.Left(4).Compare(_L8("GET "))==0)
		{
			buf = buf.Right(buf.Length()-4);
			TInt pos;
			if ((pos = buf.Locate(' '))!=KErrNotFound)
			{
				tempbuf = buf.Right(buf.Length()-(pos+1));
				if (tempbuf.Compare(_L8("HTTP/1.0\r"))==0 || tempbuf.Compare(_L8("HTTP/1.1\r"))==0 ||
					tempbuf.Compare(_L8("HTTP/1.0"))==0 || tempbuf.Compare(_L8("HTTP/1.1"))==0)
				{   
					// Format accepted
					buf = buf.Left(pos);
					if (buf.Length()>0)
					{
						delete iCurrentRequest.requestedUrl;
						iCurrentRequest.requestedUrl = buf.AllocL();
						iCurrentRequest.finished = ETrue;
					}
				}
			}
		} 
		// Format: [empty line]
		else if (buf.Length()==0 || buf.Compare(_L8("\r"))==0)
		{
			return ETrue;
		}
	}

	return EFalse;
}

void CSTTServiceSocket::DecodeRequestL()
{
	iLog->WriteLineL(_L("[TRACKER] DecodeRequest"));
	TBuf8<2048> tempbuf;
	TBuf8<20> paramname;

	paramname.Copy(_L8("info_hash"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (iCurrentRequest.infoHash!=NULL) delete iCurrentRequest.infoHash;
	iCurrentRequest.infoHash = tempbuf.AllocL();

	paramname.Copy(_L8("peer_id"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (iCurrentRequest.peerId!=NULL) delete iCurrentRequest.peerId;
	iCurrentRequest.peerId = tempbuf.AllocL();

	paramname.Copy(_L8("port"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (TLex8(tempbuf).Val(iCurrentRequest.port)!=KErrNone) iCurrentRequest.port = -1;

	paramname.Copy(_L8("uploaded"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (TLex8(tempbuf).Val(iCurrentRequest.uploaded)!=KErrNone) iCurrentRequest.uploaded = -1;

	paramname.Copy(_L8("downloaded"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (TLex8(tempbuf).Val(iCurrentRequest.downloaded)!=KErrNone) iCurrentRequest.downloaded = -1;

	paramname.Copy(_L8("left"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (TLex8(tempbuf).Val(iCurrentRequest.left)!=KErrNone) iCurrentRequest.left = -1;

	paramname.Copy(_L8("compact"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (iCurrentRequest.compact!=NULL) delete iCurrentRequest.compact;
	iCurrentRequest.compact = tempbuf.AllocL();

	paramname.Copy(_L8("event"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (tempbuf.Compare(_L8("started"))==0) iCurrentRequest.event = ESSStarted;
	else if (tempbuf.Compare(_L8("stopped"))==0) iCurrentRequest.event = ESSStopped;
	else if (tempbuf.Compare(_L8("completed"))==0) iCurrentRequest.event = ESSCompleted;
	else iCurrentRequest.event = ESSNotSepecified;

	paramname.Copy(_L8("ip"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (iCurrentRequest.ip!=NULL) delete iCurrentRequest.ip;
	iCurrentRequest.ip = tempbuf.AllocL();
	iLog->WriteL(_L("[TRACKER] Ip received: "));
	iLog->WriteLineL(tempbuf);

	paramname.Copy(_L8("numwant"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (TLex8(tempbuf).Val(iCurrentRequest.numwant)!=KErrNone) iCurrentRequest.numwant = -1;

	paramname.Copy(_L8("key"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (iCurrentRequest.key!=NULL) delete iCurrentRequest.key;
	iCurrentRequest.key = tempbuf.AllocL();

	paramname.Copy(_L8("trackerid"));
	GetParameter(*iCurrentRequest.requestedUrl,paramname,tempbuf);
	if (iCurrentRequest.trackerId!=NULL) delete iCurrentRequest.trackerId;
	iCurrentRequest.trackerId = tempbuf.AllocL();
}

// Add peer to list
void CSTTServiceSocket::AddPeer(CSTBencodedList* aPeerList, const TDesC8& aPeerId, const TDesC8& aIp, TInt aPort)
{
	iLog->WriteL(_L("[TRACKER] AddPeer: "));
	iLog->WriteL(aIp);
	iLog->WriteL(_L(":"));
	iLog->WriteLineL(aPort);
	CSTBencodedDictionary* peer = new (ELeave) CSTBencodedDictionary;

	CSTBencodedString* peerkey = CSTBencodedString::NewL(_L8("peer id"));
	CSTBencodedString* peervalue = CSTBencodedString::NewL(aPeerId);
	peer->AddEntryL(peerkey,peervalue);

	CSTBencodedString* ipkey = CSTBencodedString::NewL(_L8("ip"));
	CSTBencodedString* ipvalue = CSTBencodedString::NewL(aIp);
	peer->AddEntryL(ipkey,ipvalue);

	CSTBencodedString* portkey = CSTBencodedString::NewL(_L8("port"));
	CSTBencodedInteger* portvalue = new (ELeave) CSTBencodedInteger(aPort);
	peer->AddEntryL(portkey,portvalue);

	aPeerList->AppendL(peer);
}

CSTTTorrentListItem* CSTTServiceSocket::FindTorrent(TDesC8 &aInfoHash)
{
	TBool found = EFalse;
	CSTTBaseData* basedata = iTrackerManager->GetBaseData();
	CSTTTorrentListItem* listitem = basedata->GetTorrentListFirstItem();

	while (listitem!=NULL && !found) 
	{
		found = !((listitem->GetInfoHash()).Compare(aInfoHash));
		if (!found)
			listitem = listitem->GetNextSibling();
	}
	if (found)
		return listitem;
	else
		return NULL;
}

void CSTTServiceSocket::ComposeResponse(CSTTTorrentListItem* aListitem)
{
	iCurrentResponse.interval		= 180;
	iCurrentResponse.minInterval	= 10;
	iCurrentResponse.trackerId		= _L8("SymTorrentTracker").AllocL();

	CSTTTorrentPeerItem* peeritem;
	TInt complete=0, incomplete=0;
	peeritem = aListitem->GetPeerListFirstItem();
	while (peeritem!=NULL)
	{
		if (peeritem->Completed())
			complete++;
		else
			incomplete++;
		peeritem = peeritem->GetNextSibling();
	}
	iCurrentResponse.complete = complete;
	iCurrentResponse.incomplete = incomplete;

	iCurrentResponse.firstPeer = aListitem->GetPeerListFirstItem();
}

// Process request
void CSTTServiceSocket::ProcessRequestL()
{
	iLog->WriteLineL(_L("[TRACKER] - ProcessRequestL"));
	//iLog->WriteLineL(iCurrentRequest.ip);

	iCurrentResponse.torrentFound = EFalse;
	if (iCurrentResponse.errorMessage) delete iCurrentResponse.errorMessage;
	iCurrentResponse.errorMessage = NULL;
	if (iCurrentResponse.warningMessage) delete iCurrentResponse.warningMessage;
	iCurrentResponse.warningMessage = NULL;
	iCurrentResponse.interval = -1;
	iCurrentResponse.minInterval = -1;
	if (iCurrentResponse.trackerId) delete iCurrentResponse.trackerId;
	iCurrentResponse.trackerId = NULL;
	iCurrentResponse.complete = -1;
	iCurrentResponse.incomplete = -1;
	iCurrentResponse.firstPeer = NULL;

	if (iCurrentRequest.ip->Length()==0 || iCurrentRequest.ip->Length()>31)
	{
		TBuf<60> buf;
		TBuf8<60> buf8;
		iRemoteAddressFromSocket.Output(buf);
		buf8.Copy(buf);
		delete iCurrentRequest.ip;
		iCurrentRequest.ip = buf8.AllocL();  	
	}
	
	iLog->WriteL(_L("[TRACKER] Connected peer address: "));
	iLog->WriteLineL(*iCurrentRequest.ip);

	CSTTTorrentListItem* listitem = FindTorrent(*iCurrentRequest.infoHash);
	if (listitem != NULL)
	{
		listitem->RemoveNeedlessPeers();
		iCurrentResponse.torrentFound = ETrue;
		TInetAddr tempaddr;
		TBuf<60> tempbuf;
		tempbuf.Copy(*iCurrentRequest.ip);
		tempaddr.Input(tempbuf);
		tempaddr.SetPort(iCurrentRequest.port);

		if (iCurrentRequest.event==ESSStarted)
		{
			listitem->DeletePeerItem(tempaddr,*iCurrentRequest.peerId);
			//listitem->DeleteSimilarPeers(tempaddr,*iCurrentRequest.peerId);

			listitem->InsertPeerItemL(tempaddr,*iCurrentRequest.peerId);
			ComposeResponse(listitem);
		}
		else if (iCurrentRequest.event==ESSCompleted)
		{
			CSTTTorrentPeerItem* peer = listitem->FindPeerItem(tempaddr,*iCurrentRequest.peerId);

			// if a peer with the given peer ID and address is found then updates its status
			if (peer)
			{
				peer->SetLastCheckInTimeNow();
				peer->SetCompleted(ETrue);
				ComposeResponse(listitem);
			}
			else
			{
				// deletes peers with the same address
				CSTTTorrentPeerItem* peer = listitem->FindPeerItem(tempaddr);				
				if (peer)
					listitem->DeleteSimilarPeers(tempaddr,*iCurrentRequest.peerId);

				listitem->InsertPeerItemL(tempaddr,*iCurrentRequest.peerId);
				ComposeResponse(listitem);
			}
		}
		else if (iCurrentRequest.event==ESSStopped)
		{
			listitem->DeletePeerItem(tempaddr,*iCurrentRequest.peerId);
			//listitem->DeleteSimilarPeers(tempaddr,*iCurrentRequest.peerId);
			ComposeResponse(listitem);
		}
		else
		{
			CSTTTorrentPeerItem* peer = listitem->FindPeerItem(tempaddr,*iCurrentRequest.peerId);

			if (peer!=NULL)
			{
				peer->SetLastCheckInTimeNow();
				ComposeResponse(listitem);
			}
			else
			{
				CSTTTorrentPeerItem* tempitem2;
				TInt num = 0;
				tempitem2 = listitem->GetPeerListFirstItem();
				while (tempitem2!=NULL)
				{
					num++;
					tempitem2 = tempitem2->GetNextSibling();
				}
				num = num;


				listitem->InsertPeerItemL(tempaddr,*iCurrentRequest.peerId);
				ComposeResponse(listitem);


				num = 0;
				tempitem2 = listitem->GetPeerListFirstItem();
				while (tempitem2!=NULL)
				{
					num++;
					tempitem2 = tempitem2->GetNextSibling();
				}
				num = num;
			}
		}
	}
	else
		iCurrentResponse.torrentFound = EFalse;
}

// Generate response to the request
HBufC8* CSTTServiceSocket::GenerateResponseL()
{
	iLog->WriteLineL(_L("[TRACKER] GenerateResponseL"));
	ProcessRequestL();

	CSTBencodedDictionary* responseroot = new (ELeave) CSTBencodedDictionary;

	if (!iCurrentResponse.torrentFound)
	{
		CSTBencodedString* failkey = CSTBencodedString::NewL(_L8("failure reason"));
		CSTBencodedString* failvalue = CSTBencodedString::NewL(_L8("Torrent not found"));
		responseroot->AddEntryL(failkey,failvalue);
	}
	else
	{
		if (iCurrentResponse.errorMessage!=NULL)
		{
			CSTBencodedString* failkey = CSTBencodedString::NewL(_L8("failure reason"));
			CSTBencodedString* failvalue = CSTBencodedString::NewL(*iCurrentResponse.errorMessage);
			responseroot->AddEntryL(failkey,failvalue);
		}
		else
		{
			if (iCurrentResponse.warningMessage!=NULL)
			{
				CSTBencodedString* warnkey = CSTBencodedString::NewL(_L8("warning message"));
				CSTBencodedString* warnvalue = CSTBencodedString::NewL(*iCurrentResponse.warningMessage);
				responseroot->AddEntryL(warnkey,warnvalue);
			}
			
			if (iCurrentResponse.interval>=0)
			{
				CSTBencodedString* intkey = CSTBencodedString::NewL(_L8("interval"));
				CSTBencodedInteger* intvalue = new (ELeave) CSTBencodedInteger(iCurrentResponse.interval);
				responseroot->AddEntryL(intkey,intvalue);
			}
			
			if (iCurrentResponse.minInterval>=0)
			{
				CSTBencodedString* minikey = CSTBencodedString::NewL(_L8("min interval"));
				CSTBencodedInteger* minivalue = new (ELeave) CSTBencodedInteger(iCurrentResponse.minInterval);
				responseroot->AddEntryL(minikey,minivalue);
			}
			
			if (iCurrentResponse.trackerId!=NULL)
			{
				CSTBencodedString* trakey = CSTBencodedString::NewL(_L8("tracker id"));
				CSTBencodedString* travalue = CSTBencodedString::NewL(*iCurrentResponse.trackerId);
				responseroot->AddEntryL(trakey,travalue);
			}

			if (iCurrentResponse.complete>=0)
			{
				CSTBencodedString* compkey = CSTBencodedString::NewL(_L8("complete"));
				CSTBencodedInteger* compvalue = new (ELeave) CSTBencodedInteger(iCurrentResponse.complete);
				responseroot->AddEntryL(compkey,compvalue);
			}

			if (iCurrentResponse.incomplete>=0)
			{
				CSTBencodedString* icompkey = CSTBencodedString::NewL(_L8("incomplete"));
				CSTBencodedInteger* icompvalue = new (ELeave) CSTBencodedInteger(iCurrentResponse.incomplete);
				responseroot->AddEntryL(icompkey,icompvalue);
			}
			
			CSTBencodedString* peerskey = CSTBencodedString::NewL(_L8("peers"));
			CSTBencodedList* peersvalue = new (ELeave) CSTBencodedList;
			responseroot->AddEntryL(peerskey,peersvalue);

			// add peers to peersvalue
			TInt num=0;
			CSTTTorrentPeerItem* peeritem = iCurrentResponse.firstPeer;

			while (peeritem!=NULL && (num<iCurrentRequest.numwant || iCurrentRequest.numwant<0))
			{
				AddPeer(peersvalue,*(peeritem->PeerId()),*(peeritem->PeerIp()),peeritem->Port());
				iLog->WriteL(_L("[TRACKER] Response sent: "));
				iLog->WriteL(*peeritem->PeerIp());
				iLog->WriteL(_L(":"));
				iLog->WriteLineL(peeritem->Port());
				
				peeritem = peeritem->GetNextSibling();
				num++;
			}
		}
	}
	
	return responseroot->BencodeL();
}

void CSTTServiceSocket::SendRequestedInformationL()
{
	iState = ESSPrepareSending;
	if (iCurrentRequest.finished)
	{
		TBuf<120> tempbuf;
		// start service
		DecodeRequestL();
		
		iResponse = GenerateResponseL();

		TBuf8<1048> msg;
		msg.Copy(_L8(""));
		msg.Append(_L8("HTTP/1.0 200 OK\n"));
		msg.Append(_L8("Content-type: text/plain\n"));
		msg.Append(_L8("Content-length: "));
		msg.AppendNum((*iResponse).Length());
		msg.Append(_L8("\n"));
		msg.Append(_L8("\n"));

		iResponseHeader = msg.AllocL();

		iState = ESSSendingResponseHeader;
		iSocket.Write(*iResponseHeader,iStatus);
		SetActive();
	}
	else
	{
		ReadIncoming();
	}
}

void CSTTServiceSocket::SendRequestedContentL()
{
	delete iResponseHeader;

	iState = ESSSendingResponse;
	iSocket.Write(*iResponse,iStatus);
	SetActive();
}

void CSTTServiceSocket::AfterSending()
{
	delete iResponse;

	if (iTemporaryBuffer.Length()==0)
	{
		TransferDisconnect();
	}
	else
	{
		iState = ESSProcessIncoming;

		iCurrentRequest.finished = EFalse;
		delete iCurrentRequest.requestedUrl;
		iCurrentRequest.requestedUrl = _L8("").AllocL();

		if (!ProcessIncomingL())
			ContinueReading();
		else
			SendRequestedInformationL();
	}
}

/*
void CSTTServiceSocket::DecodeLocation(TDes16 &aWwwRoot, TDes8 &aLocation, TDes16 &aResult)
{
	TBuf8<1048> temp;
	temp.Copy(aLocation);
	TInt i;

	for (i=0;i<temp.Length();i++)
		if (temp[i]=='/')
			temp[i]='\\';

	aResult.Copy(aWwwRoot);
	MyAppend16_8(aResult,temp);
}
*/


// Other methods --------------------------------------------------------------

void CSTTServiceSocket::MyAppend16_8(TDes16 &aTarget,TDesC8 &aSource)
{
	TInt i;
	for (i=0;i<aSource.Length();i++)
	{
		aTarget.Append(aSource[i]);
	}
}

// Encode Url
void CSTTServiceSocket::EncodeUrl(TDesC8 &aData, TDes8 &aResult)
{
	TInt i;

	for (i=0;i<aData.Length();i++)
	{
		TChar actChar = aData[i];
		TUint charCode = actChar;

		if (actChar==' ')
			aResult.Append('+');
		else if (!( (47<charCode && charCode<58) || (64<charCode && charCode<91) || (96<charCode && charCode<123) ))
		{
			aResult.Append('%');
			TBuf<4> Hex;
			Hex.Num(charCode, EHex);			
			if (charCode<16)
				aResult.Append('0');
			aResult.Append(Hex.Left(2));
		}
		else
			aResult.Append(actChar);
	}
}

// Decode Url
void CSTTServiceSocket::DecodeUrl(TDesC8 &aData, TDes8 &aResult)
{
	TInt i;

	for (i=0;i<aData.Length();i++)
	{
		TChar actChar = aData[i];

		if (actChar=='+')
			aResult.Append(' ');
		else if (actChar=='%')
		{
			TBuf8<2> Hex;
			TUint charCode;

			Hex.Append(aData[i+1]);
			Hex.Append(aData[i+2]);
			TLex8(Hex).Val(charCode, EHex);
			
			TChar actChar = charCode;
			aResult.Append(actChar);		
			i+=2;						
		}
		else
			aResult.Append(actChar);
	}
}

// Returns the needed parameter value
void CSTTServiceSocket::GetParameter(TDesC8 &aUrl, TDesC8 &aParameter, TDes8 &aValue)
{	
	TInt pos = aUrl.Locate('?');
	TBuf8<1024> paramDatas;
	paramDatas.Append(aUrl.Mid(pos+1));

	TBool search = ETrue;
	aValue.Copy(_L8(""));

	while (search)
	{
		pos = paramDatas.Locate('&');		
		TBuf8<1024> paramValuePair;

		if (pos == KErrNotFound)
		{
			search = EFalse;
			paramValuePair = paramDatas;
			aValue.Append(_L(""));
		}
		else
		{
			paramValuePair = paramDatas.Left(pos);
			paramDatas = paramDatas.Mid(pos+1);
		}

		TInt valuePos = paramValuePair.Locate('=');

		TBuf8<1024> param = paramValuePair.Left(valuePos);

		TBool eq = ETrue;

		for (TInt i=0; i<aParameter.Length(); i++)
		{
			if (aParameter[i] != param[i])
			{
				eq = EFalse;
				break;
			}
		}

		if (eq)
		{
			search = EFalse;
			TBuf8<1024> value;
			value.Append(paramValuePair.Mid(valuePos+1));
			TBuf8<1024> decodedValue;
			DecodeUrl(value,decodedValue);
			aValue.Append(decodedValue);
		}
	}
}

// Active object methods ------------------------------------------------------

void CSTTServiceSocket::DoCancel()
{
	iSocket.CancelAll();
	// cancel all asychronous stuff
}

void CSTTServiceSocket::RunL()
{
    switch (iStatus.Int())
    {
        case KErrNone:
			switch (iState)
			{
				case ESSWaitingForIncoming: 
					ReadIncoming();
					break;
				case ESSSendingResponseHeader: 
					SendRequestedContentL();
					break;
				case ESSSendingResponse: 
					AfterSending();
					break;
				default:
					break;
			}
			break;
		default:
			switch (iState)
			{
				default:
					TransferDisconnect();
					iState = ESSFinished;
					break;
			}
			break;
	}
}

/* ************************************************************************* */
